/*
 * Copyright (c) 2010 Aalto University and RWTH Aachen University.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * @file
 * This file provides a framework for modifying HIP packets. It includes
 * adding new parameters in the correct order and adapting the various
 * headers.
 *
 * @brief Framework for the midauth extensions
 *
 * @author Thomas Jansen
 * @author Dominic Gatzen
 * @author Christof Mroz <christof.mroz@rwth-aachen.de>
 *
 * @note: According to draft-heer-hip-middle-auth-00 we SHOULD support IP-level
 * fragmentation for IPv6 and MUST support IP-level fragmentation for IPv4.
 * Currently we do neither.
 */

#define _BSD_SOURCE

#include <openssl/rand.h>
#include <stdbool.h>
#include <string.h>

#include "lib/core/builder.h"
#include "lib/core/debug.h"
#include "lib/core/solve.h"
#include "modules/midauth/hipd/midauth.h"
#include "modules/midauth/lib/midauth_builder.h"
#include "firewall_defines.h"
#include "rewrite.h"
#include "midauth.h"


static const uint8_t DEFAULT_DIFFICULTY = 1;

// runtime configuration
bool use_midauth = false;

/**
 * Initialize the midauth extension for the firewall.
 */
void midauth_init(void)
{
    HIP_DEBUG("== midauth enabled ==\n");
}

/**
 * Get the lifetime of a midauth challenge.
 *
 * @param difficulty The difficulty of the puzzle.
 * @return The lifetime of the challenge.
 *
 * @todo Lifetime should depend on difficulty.
 */
static inline uint8_t lifetime(UNUSED const uint8_t difficulty)
{
    return 2;
}

/**
 * Add a CHELLENGE_REQUEST parameter to a HIP packet passing through the
 * firewall.
 *
 * @param ctx The packet context.
 * @param[out] midauth_nonce The nonce added by the firewall for this handshake.
 * @return one on success, zero otherwise
 */
int hipfw_midauth_add_challenge(struct hip_fw_context *const ctx,
                                uint8_t midauth_nonce[MIDAUTH_DEFAULT_NONCE_LENGTH])
{
    struct hip_challenge_request request;
    static const size_t          min_length = sizeof(request)
                                              - sizeof(request.tlv)
                                              - sizeof(request.opaque);

    if (use_midauth) {
        HIP_ASSERT(midauth_nonce);
        HIP_ASSERT(ctx->packet_type == HIP_PACKET);

        /* note: the length cannot be calculated with calc_param_len() */
        hip_set_param_contents_len(&request.tlv,
                                   min_length + MIDAUTH_DEFAULT_NONCE_LENGTH);
        hip_set_param_type(&request.tlv, HIP_PARAM_CHALLENGE_REQUEST);

        request.K        = DEFAULT_DIFFICULTY;
        request.lifetime = lifetime(DEFAULT_DIFFICULTY);

        if (!RAND_bytes(request.opaque, MIDAUTH_DEFAULT_NONCE_LENGTH)) {
            HIP_ERROR("Failed to generate CHALLENGE_REQUEST nonce\n");
            return 0;
        }

        // IP (and UDP, if needed) frames will be updated upon send
        // Implicitly calls hip_fw_context_enable_write() if needed
        if (!hipfw_splice_param(ctx, &request.tlv)) {
            HIP_ERROR("Failed to splice CHALLENGE_REQUEST into existing packet\n");
            return 0;
        }

        /* store nonce per-connection per-peer */
        memcpy(midauth_nonce, request.opaque, MIDAUTH_DEFAULT_NONCE_LENGTH);
    }

    return 1;
}

/**
 * Getter for the length of the opaque field in a CHALLENGE_REPSONSE parameter.
 *
 * @param response The CHALLENGE_REPSONSE parameter.
 * @return The length of the opaque value.
 */
static uint8_t hip_challenge_response_opaque_len(const struct hip_challenge_response *response)
{
    static const size_t min_len = sizeof(*response) -
                                  sizeof(response->tlv) -
                                  sizeof(response->opaque);

    return hip_get_param_contents_len(&response->tlv) - min_len;
}

/**
 * Helper function for verifying a CHALLENGE_RESPONSE parameter.
 *
 * @param ctx The packet context.
 * @param response The CHALLENGE_RESPONSE parameter.
 * @param midauth_nonce The stored nonce for this association.
 * @return ERROR on validation error,
 *         RESPONSE_CORRECT if validation was successful,
 *         RESPONSE_INCORRECT if nonces match but solution is incorrect, and
 *         RESPONSE_NO_MATCH if no matching nonces are stored.
 */
static enum {
    ERROR,
    RESPONSE_CORRECT,
    RESPONSE_INCORRECT,
    RESPONSE_NO_MATCH
}
verify_response(const struct hip_fw_context *const ctx,
                const struct hip_challenge_response *const response,
                const uint8_t midauth_nonce[MIDAUTH_DEFAULT_NONCE_LENGTH])
{
    //
    // TODO: - check lifetime
    //       - compare with connection state entry (K, nonce)
    //       - use HITs from packet or from ctx?
    //

    const uint8_t len = hip_challenge_response_opaque_len(response);
    const bool match  = (len == MIDAUTH_DEFAULT_NONCE_LENGTH) &&
                        (memcmp(response->opaque, midauth_nonce,
                                MIDAUTH_DEFAULT_NONCE_LENGTH) == 0);

    if (!match) {
        return RESPONSE_NO_MATCH;
    }

    const struct hip_common *const hip = ctx->transport_hdr.hip;

    if (response->K > 0) {
        struct puzzle_hash_input tmp_puzzle;

        if (hip_midauth_puzzle_seed(response->opaque, len, tmp_puzzle.puzzle)) {
            HIP_ERROR("failed to derive midauth puzzle\n");
            return ERROR;
        }
        tmp_puzzle.initiator_hit = hip->hits;
        tmp_puzzle.responder_hit = hip->hitr;
        memcpy(tmp_puzzle.solution, response->J, PUZZLE_LENGTH);

        if (hip_verify_puzzle_solution(&tmp_puzzle, response->K)) {
            HIP_ERROR("Incorrect CHALLENGE_RESPONSE found\n");
            return RESPONSE_INCORRECT;
        }
    }

    return RESPONSE_CORRECT;
}

/**
 * Verify a puzzle solution in the CHALLENGE_RESPONSE parameter.
 *
 * @param ctx The packet context.
 * @param midauth_nonce The stored nonce for this association.
 * @return one on success, zero otherwise
 */
int hipfw_midauth_verify_challenge(const struct hip_fw_context *const ctx,
                                   const uint8_t midauth_nonce[MIDAUTH_DEFAULT_NONCE_LENGTH])
{
    const struct hip_challenge_response *response;

    if (use_midauth) {
        HIP_ASSERT(ctx->packet_type == HIP_PACKET);

        response = hip_get_param(ctx->transport_hdr.hip,
                                 HIP_PARAM_CHALLENGE_RESPONSE);
        if (!response) {
            HIP_ERROR("Challenge response expected but not found\n");
            return 0;
        }

        do {
            switch (verify_response(ctx, response, midauth_nonce)) {
            case RESPONSE_CORRECT:
                HIP_DEBUG("Correct CHALLENGE_RESPONSE found\n");
                return 1;

            case RESPONSE_INCORRECT:
            case RESPONSE_NO_MATCH:
                break;

            default:
                HIP_ERROR("Unable to compute challenge verification.\n");
            }

            response = (const struct hip_challenge_response *)
                       hip_get_next_param(ctx->transport_hdr.hip, &response->tlv);
        } while (response &&
                 hip_get_param_type(response) == HIP_PARAM_CHALLENGE_RESPONSE);

        HIP_ERROR("CHALLENGE_RESPONSE found, but no matching nonce\n");
        return 0;
    }

    return 1;
}
