/*
 * Copyright (c) 2010 Aalto University and RWTH Aachen University.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * @file
 * This file contains PISA specific functionality for the firewall.
 *
 * @author Henrik Ziegeldorf
 */

#define _BSD_SOURCE

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/x509.h>

#include "firewall_defines.h"
#include "conntrack.h"
#include "lib/core/common.h"
#include "lib/core/ife.h"
#include "lib/core/builder.h"
#include "lib/core/debug.h"
#include "lib/core/cert.h"

#include "pisa.h"

// runtime configuration
int         use_pisa              = false;
const char *root_certificate_file = NULL;

static STACK_OF(X509) * root_chain = NULL;

/**
 * Init function for PISA functionality.
 * This functions preloads the pisa root certificate, which is used
 * to verify certificates presented by PISA trust points.
 *
 * @return 0 on success, negative if an error occurred
 */
int pisa_init(void)
{
    int   err  = 0;
    X509 *cert = NULL;

    if (use_pisa) {
        /* Preload root certifcate of community operator. */
        if (!root_certificate_file) {
            HIP_DEBUG("Using default trust root at: " HIPL_SYSCONFDIR "/pisa-root-cert.pem\n");
            root_certificate_file = HIPL_SYSCONFDIR "/pisa-root-cert.pem";
        }
        HIP_IFEL(!(cert = cert_load_x509_certificate(root_certificate_file)),
                 -1, "Could not load certificate of community operator at: %s \n", root_certificate_file);
        HIP_IFEL(!(root_chain = sk_X509_new_null()),
                 -1, "Memory allocation failure.\n");
        sk_X509_push(root_chain, cert);

        HIP_DEBUG("PISA activated\n");
    }

out_err:
    return err;
}

/**
 * Uninit function for PISA functionality.
 *
 * @return 0
 */
int pisa_uninit(void)
{
    if (root_chain) {
        sk_X509_free(root_chain);
    }
    return 0;
}

/**
 * Helper function that converts the special RSA and DSA key structures
 * to the generic EVP_PKEY structure.
 *
 * @param key   the RSA or DSA key structure
 * @param algo  either HIP_HI_RSA or HIP_HI_DSA
 *
 * @return      the EVP_PKEY structure that wraps the original key,
 *              or NULL on error
 */
static EVP_PKEY *any_key_to_evp_key(void *key, int algo)
{
    int       err = 0;
    EVP_PKEY *ret = NULL;

    HIP_IFEL(!(ret = EVP_PKEY_new()), -1, "Could not init EVP_PKEY wrapper\n");

    switch (algo) {
    case HIP_HI_RSA:
        err = EVP_PKEY_assign_RSA(ret, key);
        break;
    case HIP_HI_DSA:
        err = EVP_PKEY_assign_DSA(ret, key);
        break;
    default:
        HIP_DEBUG("Unknown algorithm \n");
    }
    HIP_IFEL(err == 0, -1, "Could not assign key to EVP_PKEY.\n");

    return ret;

out_err:
    EVP_PKEY_free(ret);
    return NULL;
}

/**
 * The R2 packet contains the certificate of the trust point in PISA.
 * The functions extracts the certificate, matches the contained public
 * key against the HI provided in the R1 and tries to build and verify
 * a certificate chain.
 *
 * For an update exchange, a certificate must be contained either in the U2
 * (if the exchange was started by the Initiator) or in the U3
 * (if the exchange was started by the Responder). If a certificate cannot be
 * found in these situations an error is returned.
 *
 * @param common    the R2 or U2 packet
 * @param tuple     the connection tracking tuple
 * @param ctx       the firewall context
 * @return          0 on success, negative error code otherwise
 */
int pisa_handle_certificate(struct hip_common *const common,
                            UNUSED struct tuple *const tuple,
                            UNUSED const struct hip_fw_context *const ctx)
{
    int       err  = 0;
    X509     *cert = NULL;
    EVP_PKEY *pkey = NULL;

    if (use_pisa) {
        /* Should there be a certificate?
         * Not, if this update is not sent by the Responder. */
        if (!(tuple->direction == REPLY_DIR)) {
            return 0;
        }

        /* Extract certificate of trust point from the packet. */
        if (!(cert = cert_get_X509_from_msg(common))) {
            HIP_DEBUG("Could not find trust-point certificate in R2/U2\n");
            return -1;
        }

        /* Match HI against public key in given certificate. */
        pkey = any_key_to_evp_key(tuple->hip_tuple->data->src_pub_key,
                                  hip_get_host_id_algo(tuple->hip_tuple->data->src_hi));
        HIP_IFEL(!cert_match_public_key(cert, pkey),
                 -1, "HI does not match public key in given certificate\n");
        HIP_DEBUG("HI matchtes given certificate.\n");

        /* Check certificate of trust point. */
        HIP_IFEL(cert_verify_chain(cert, NULL, root_chain, NULL),
                 -1, "Could not verify trust point certificate.\n");

        HIP_DEBUG("Verified certificate of trust-point.\n");
    }

out_err:
    return err;
}
