#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <openssl/x509.h>
#include <openssl/pem.h>

#include "builder.h"
#include "debug.h"
#include "ife.h"
#include "common.h"
#include "cert.h"

/**
 * This function encodes the given certificate to its DER encoding for
 * transmission over the wire. The encoded certificate is written to @ *buf.
 *
 * @NOTE: The encoded data is in binary form and may contain embedded zeroes.
 * Functions such as strlen() will not return the correct length of the encoded structure.
 * Therefore length value returned by this function should always be used.
 *
 * @param cert      the certificate to encode
 * @param buf       the output is written here
 *
 * @return          the length of the encoded data
 */
int cert_X509_to_DER(X509 *const cert, unsigned char **buf)
{
    int len;

    if (!cert) {
        HIP_ERROR("Cannot encode NULL-certificate\n");
        return -1;
    }
    if (!buf) {
        HIP_ERROR("Cannot create output buffer at NULL-pointer.\n");
        return -1;
    }

    *buf = NULL;
    len  = i2d_X509(cert, buf);

    if (len < 0) {
        HIP_ERROR("Could not DER-encode the given certificate.\n");
        return -1;
    }
    return len;
}

/**
 * Function to decode a DER-encoded certificate to the internal openssl X509
 * structure.
 *
 * @param buf   the buffer from which the DER-encoded certificate is read
 * @param len   the length of the DER-encoded certificate
 *              (NOTE: strlen() and similar functions fail, since DER is basically
 *               binary data that can contain 0-bytes).
 *
 * @return      the openssl X509 structure corresponding the DER-encoded certificate,
 *              NULL if errors occured
 */
X509 *cert_DER_to_X509(const unsigned char *const buf, const int len)
{
    const unsigned char *p;

    if (!buf) {
        HIP_ERROR("Cannot decode from NULL-buffer\n");
        return NULL;
    }
    if (len <= 0) {
        HIP_ERROR("Cannot decode certificate of length <= 0\n");
        return NULL;
    }
    p = buf;
    return d2i_X509(NULL, (const unsigned char **)  &p, len);
}

/**
 * Load a X509 certificate from @ file. If @ file contains more
 * than one certificate, the certificate at the top of the file is
 * returned.
 *
 * @file    the file to load the certficate from
 * @return  a X509 certificate on success, NULL on error
 */
X509 *cert_load_x509_certificate(const char *const file)
{
    int   err  = 0;
    FILE *fp   = NULL;
    X509 *cert = NULL;

    HIP_IFEL(!file, -1, "Cannot read certificate from NULL-filename.\n");
    fp = fopen(file, "rb");
    HIP_IFEL(!fp, -1, "Could not open file for reading: %s\n", file);
    cert = PEM_read_X509(fp, NULL, NULL, NULL);
    HIP_IFEL(fclose(fp), -1, "Error closing file: %s\n", file);

    return cert;

out_err:
    X509_free(cert);
    return NULL;
}

/**
 * This functions looks for hip_cert parameter in @ msg and tries to
 * decode the data in the first certificate parameter to a X509 certificate.
 *
 * @param msg   the message to extract the certificate from
 * @return      the first X509 certificate found in the message on success,
 *              NULL on error and if no certificates were found
 */
X509 *cert_get_X509_from_msg(struct hip_common *const msg)
{
    const struct hip_cert *param_cert = NULL;

    if (!(param_cert = hip_get_param(msg, HIP_PARAM_CERT))) {
        HIP_ERROR("Message contains no certificate.\n");
        return NULL;
    }

    /* The contents of the certificate begin after the header of the hip_cert parameter */
    return cert_DER_to_X509((const unsigned char *) (param_cert + 1),
                            ntohs(param_cert->length) - sizeof(struct hip_cert) + sizeof(struct hip_tlv_common));
}

/**
 * Compare a given public key @ pkey with the public key
 * contained in @ cert.
 *
 * @param cert  the X509 certificate
 * @param pkey  the public key to match
 *
 * @return 1 if match, 0 otherwise
 */
int cert_match_public_key(X509 *cert, const EVP_PKEY *pkey)
{
    EVP_PKEY *pkey2 = NULL;

    if (!cert) {
        return 0;
    }
    if (!pkey) {
        return 0;
    }
    pkey2 = X509_get_pubkey(cert);
    return EVP_PKEY_cmp(pkey2, pkey);
}

/**
 * Build and verify a certificate chain.
 *
 * @param leaf_cert             the certificate to verify
 * @param trusted_lookup_dir    certificates in this directory are used as root certificates
 * @param trusted_chain         a certificate stack, that can contain additional trusted certificate
 * @param untruste_chain        a chain of untrusted certificates, that can be used to build a complete certificate chain
 *
 * @return                      0 if a certificate chain could be built and verified, a non-zero error code otherwise
 */
int cert_verify_chain(X509 *leaf_cert,
                      const char *trusted_lookup_dir,
                      STACK_OF(X509) *trusted_chain,
                      STACK_OF(X509) *untrusted_chain)
{
    int             err              = 0;
    X509_LOOKUP    *lookup           = NULL;
    X509_STORE     *verify_ctx_store = NULL;
    X509_STORE_CTX *verify_ctx       = NULL;

    if (!leaf_cert) {
        HIP_ERROR("Cannot verify NULL-certificate.\n");
        return -1;
    }

    if (!trusted_lookup_dir && !trusted_chain) {
        HIP_ERROR("Need trusted dir or trusted chain.\n");
        return -1;
    }

    /* Build the verify context */
    HIP_IFEL(!(verify_ctx_store = X509_STORE_new()),
             -1, "Failed to init certificate store \n");
    HIP_IFEL(!(lookup = X509_STORE_add_lookup(verify_ctx_store, X509_LOOKUP_hash_dir())),
             -1, "Failed to init lookup directory \n");
    if (trusted_lookup_dir) {
        HIP_IFEL(!X509_LOOKUP_add_dir(lookup, trusted_lookup_dir, X509_FILETYPE_PEM),
                 -1, "Failed to add directory %s to trusted lookup resources \n", trusted_lookup_dir);
    } else {
        X509_LOOKUP_add_dir(lookup, NULL, X509_FILETYPE_DEFAULT);
    }

    HIP_IFEL(!(verify_ctx = X509_STORE_CTX_new()),
             -1, "Failed to allocate new verify context \n");
    HIP_IFEL(!X509_STORE_CTX_init(verify_ctx, verify_ctx_store, leaf_cert, untrusted_chain),
             -1, "Failed to setup verify context\n");
    if (trusted_chain) {
        X509_STORE_CTX_trusted_stack(verify_ctx, trusted_chain);
    }

    /* Finally do the verification and output some info on error */
    OpenSSL_add_all_algorithms();
    err = X509_verify_cert(verify_ctx);

    if (err != 1) {
        err = X509_STORE_CTX_get_error(verify_ctx);
        HIP_DEBUG("X509 verify cert error: %d \n", err);
        HIP_DEBUG("at depth: %d \n", X509_STORE_CTX_get_error_depth(verify_ctx));
        HIP_DEBUG("reason: %s\n", X509_verify_cert_error_string(err));
        HIP_DEBUG("certificate:\n");
        X509_print_fp(stderr, X509_STORE_CTX_get_current_cert(verify_ctx));
    } else {
        err = 0;
    }

out_err:
    X509_STORE_CTX_free(verify_ctx);
    return err;
}
