/*
 * Copyright (c) 2010 Aalto University and RWTH Aachen University.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * @file
 * This file contains functions that are specific to PISA.
 *
 * @author Thomas Jansen
 * @author Henrik Ziegeldorf
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hipd/hipd.h"
#include "lib/core/builder.h"
#include "lib/core/debug.h"
#include "lib/core/ife.h"
#include "lib/core/cert.h"
#include "hipd/pkt_handling.h"
#include "modules/update/hipd/update.h"
#include "modules/midauth/hipd/midauth.h"

#include "pisa.h"

static X509 *midauth_cert = NULL;

/**
 *  Handler that adds the PISA trust point certificate in an R2 message.
 */
static int hip_add_certificate(struct hip_common *msg, X509 *certificate)
{
    int            err = 0;
    unsigned char *buf;
    int            len = 0;

    /* Sanity checks */
    HIP_IFEL(!msg, -1, "Message is NULL\n");
    HIP_IFEL(!certificate, -1, "Certificate is NULL\n");

    /* Encode the certificate to DER and build the certificate parameter. */
    HIP_IFEL((len = cert_X509_to_DER(certificate, &buf)) < 0,
             -1, "Encoding error\n");
    HIP_IFEL(hip_build_param_cert(msg, 0, 1, 1, HIP_CERT_X509V3, buf, len),
             -1, "Building of certificate parameter failed.\n");

    return 0;

out_err:
    return err;
}

/**
 *  Handler that adds the PISA trust point certificate in an R2 message.
 *
 *  @NOTE:  The certificate is regarded non-critical thus the function does
 *          not fail even if no midauth certificate is available.
 */
static int hip_add_certificate_r2(UNUSED const uint8_t packet_type,
                                  UNUSED const uint32_t ha_state,
                                  struct hip_packet_context *ctx)
{
    if (hip_add_certificate(ctx->output_msg, midauth_cert)) {
        HIP_DEBUG("Sending R2 without midauth certificate.\n");
    }
    return 0;
}

/**
 *  Handler that adds the PISA trust point certificate to the second or third update message.
 *  A certificate should only be included if the previous update packet containted
 *  a middlebox challenge.
 *
 *  @NOTE:  The certificate is regarded non-critical thus the function does
 *          not fail even if no midauth certificate is available.
 */
static int hip_add_certificate_update(UNUSED const uint8_t packet_type,
                                      UNUSED const uint32_t ha_state,
                                      struct hip_packet_context *ctx)
{
    if (!midauth_cert) {
        HIP_DEBUG("No midauth certificate available.\n");
        return 0;
    }

    /* Include a midauth certificate in the U2 or U3, if available. */
    if (hip_classify_update_type(ctx->input_msg) == FIRST_UPDATE_PACKET ||
        hip_classify_update_type(ctx->input_msg) == SECOND_UPDATE_PACKET) {
        if (hip_get_param_contents(ctx->input_msg, HIP_PARAM_CHALLENGE_REQUEST)) {
            if (hip_add_certificate(ctx->output_msg, midauth_cert)) {
                HIP_ERROR("Failed to add midauth certificate to update message.\n");
                return -1;
            }
        } else {
            HIP_DEBUG("No middlebox found in previous update, omitting midauth certificate.\n");
        }
    }

    return 0;
}

/**
 * Initialize PISA functionality in the HIPD.
 * This functions registers handlers to add the trust point certificate in R2 and U2 messages.
 */
int hip_pisa_init(void)
{
    int err = 0;

    /* Add midauth certificate to R2 packet */
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_NONE, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_UNASSOCIATED, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_I1_SENT, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_I2_SENT, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_R2_SENT, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_I2, HIP_STATE_ESTABLISHED, &hip_add_certificate_r2, 40500),
             -1, "Error on registering PISA handle function.\n");

    /* Add midauth certificate to UPDATE packet */
    HIP_IFEL(hip_register_handle_function(HIP_UPDATE, HIP_STATE_ESTABLISHED, &hip_add_certificate_update, 20752),
             -1, "Error on registering PISA handle function.\n");
    HIP_IFEL(hip_register_handle_function(HIP_UPDATE, HIP_STATE_R2_SENT, &hip_add_certificate_update, 20752),
             -1, "Error on registering PISA handle function.\n");

    /* Preload the midauth certificate. */
    if (!(midauth_cert = cert_load_x509_certificate(HIPL_SYSCONFDIR "/pisa_cert.pem"))) {
        HIP_DEBUG("Could not load PISA trust point certificate. (critical only, if host acts as PISA trust point)\n");
    }

out_err:
    return err;
}
